/*
Copyright (c) 1991-1992, University of Illinois board of trustees.  All
rights reserved.  Written by Michael Maciukenas at the Ribosomal Database
Project.  Design and implementation guidance by Niels Larsen, Gary Olsen,
Carl Woese.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

typedef struct DYNAMIC {
	unsigned length;		/* length of string */
	unsigned alloc;			/* memory allocated for string */
	char *data;				/* string data */
	} dynamic;

#define BLOCKSIZE 256

char *new_string()
/* allocates a new dynamic string structure */
{
	dynamic *tmp;

	tmp=(dynamic *)malloc(sizeof(dynamic));
	if(tmp==NULL)
		return(NULL);
	tmp->length=0;
	tmp->data=(char *)malloc(BLOCKSIZE+1);
	if(tmp->data==NULL)
	{
		free(tmp);
		return(NULL);
	}
	tmp->data[0]='\0';
	tmp->alloc=BLOCKSIZE;
	return((char *)tmp);
}

char *dstring(string)
char *string;
/* returns a pointer to the string itself.  DO NOT MODIFY IT'S LENGTH! */
{
	dynamic *tmp;

	tmp=(dynamic *)string;
	return(tmp->data);
}

int dlength(string)
char *string;
/* returns the length of the given string */
{
	dynamic *tmp;

	tmp=(dynamic *)string;
	return(tmp->length);
}

int del_string(string, loc, len)
/* deletes len characters starting at loc
** returns 1 of ok, 0 if no memory available */
char *string;
int len, loc;
{
	dynamic *tmp;
	char *check;
	char *a,*b, *c;

	if(len==0)
		return(1);
	tmp=(dynamic *)string;
	if(loc+len>tmp->length)
		len=tmp->length-loc;
	/* move end of string */
	a= &tmp->data[loc];
	b= &tmp->data[loc+len];
	c= &tmp->data[tmp->length];
	while(b<=c)
		*(a++)= *(b++);

	tmp->length-=len;
	return(1);
}

int ins_string(string, loc, str)
/* inserts a string at the desired location, allocating extra memory if needed.
** returns 1 of ok, 0 if no memory available */
char *string;
char *str;
int loc;
{
	dynamic *tmp;
	char *check;
	char *a,*b, *c;
	int len=strlen(str);

	if(len==0)
		return(1);
	tmp=(dynamic *)string;
	while(tmp->length+len>tmp->alloc)
	{
		check=(char *)realloc(tmp->data, tmp->alloc+BLOCKSIZE+1);
		if(check==NULL)
			return(0);
		else
		{
			tmp->data=check;
			tmp->alloc+=BLOCKSIZE;
		}
	}
	/* move end of string */
	a= &tmp->data[tmp->length];
	b= &tmp->data[tmp->length+len];
	c= &tmp->data[loc];
	while(a>=c)
		*(b--)= *(a--);

	strncpy(&tmp->data[loc], str, len);
	tmp->length+=len;
	return(1);
}

int ins_char(string, loc, ch)
/* inserts a character at the desired location, allocating extra memory if
	needed.  returns 1 of ok, 0 if no memory available */
char *string;
int ch;
int loc;
{
	dynamic *tmp;
	char *check;
	char *a,*b, *c;

	tmp=(dynamic *)string;
	if(tmp->length==tmp->alloc)
	{
		check=(char *)realloc(tmp->data, tmp->alloc+BLOCKSIZE+1);
		if(check==NULL)
			return(0);
		else
		{
			tmp->data=check;
			tmp->alloc+=BLOCKSIZE;
		}
	}
	/* move end of string */
	a= &tmp->data[tmp->length];
	b= &tmp->data[tmp->length+1];
	c= &tmp->data[loc];
	while(a>=c)
		*(b--)= *(a--);

	tmp->data[loc]=ch;
	tmp->length+=1;
	return(1);
}

int add_string(string, str)
/* adds a string to the dynamic string, allocating extra memory if needed.
** returns 1 of ok, 0 if no memory available */
char *string;
char *str;
{
	dynamic *tmp;
	char *check;

	if(strlen(str)==0)
		return(1);
	tmp=(dynamic *)string;
	while(tmp->length+strlen(str)>tmp->alloc)
	{
		check=(char *)realloc(tmp->data, tmp->alloc+BLOCKSIZE+1);
		if(check==NULL)
			return(0);
		else
		{
			tmp->data=check;
			tmp->alloc+=BLOCKSIZE;
		}
	}
	strcpy(&tmp->data[tmp->length], str);
	tmp->length+=strlen(str);
	return(1);
}

int add_char(string, ch)
/* adds a byte to the dynamic string, allocating extra memory if needed.
** returns 1 of ok, 0 if no memory available */
char *string;
int ch;
{
	dynamic *tmp;
	char *check;

	tmp=(dynamic *)string;
	if(tmp->length==tmp->alloc)
	{
		check=(char *)realloc(tmp->data, tmp->alloc+BLOCKSIZE+1);
		if(check==NULL)
			return(0);
		else
		{
			tmp->data=check;
			tmp->alloc+=BLOCKSIZE;
		}
	}
	tmp->data[tmp->length++]=ch;
	tmp->data[tmp->length]='\0';
	return(1);
}

char *freeze_string(string)
/* pulls the string data out of the dynamic string and makes a static string
** out of it.  returns NULL if there is no memory for the operation */
char *string;
{
	dynamic *tmp;

	tmp=(dynamic *)string;
	string=tmp->data;
	free(tmp);
	return(string);
}

typedef struct LIST *list;

typedef struct LIST {
/* a list of objects */
	list obj;		/* the data item, most likely a pointer of some
						type */
	list next;
	list prev;
	} listd;

/* notes:	root->obj!=NULL */

list newlist()
/* return a new list */
{
	list tmp;

	tmp=(list)malloc(sizeof(listd));
	if(tmp!=NULL)
	{
		tmp->next=NULL;
		tmp->prev=NULL;
		tmp->obj=tmp;
	}
	return(tmp);
}

list addnode(root, obj)
/* adds the object to the list after the current node */
/* returns root if ok, 0 if not */
list root;
void *obj;
{
	list tmp;
	list node=root->obj;

	tmp=(list)malloc(sizeof(listd));
	if(tmp!=NULL)
	{
		tmp->obj=obj;
		tmp->next=node->next;
		tmp->prev=node;
		node->next=tmp;
		if(tmp->next!=NULL)
			tmp->next->prev=tmp;
		root->obj=tmp;
		return(root);
	}
	else
		return((list)0);
}

list findnode(root, obj)
/* finds the object in the list, and sets it to be the current node */
/* returns root if ok, 0 if not */
list root;
void *obj;
{
	list node=root->next;

	for(;node!=NULL;node=node->next)
		if(node->obj==obj)
			break;
	if(node!=NULL)
	{
		root->obj=node;
		return(root);
	}
	else
		return((list)0);
}

list rmnode(n)
/* removes the node n from the list */
list n;
{
	list r;

	if(n->prev==NULL)
		return(n);
	
	n->prev->next=n->next;
	if(n->next!=NULL)
		n->next->prev=n->prev;
	r=n->prev;

	free(n);

	return(r);
}

list rmcurr(root)
/* removes the current node from the list */
list root;
{
	list node=root->obj;

	if(node!=root)
	{
		root->obj=node->next;
		if(root->obj==NULL)
			root->obj=node->prev;

		node->prev->next=node->next;
		if(node->next!=NULL)
			node->next->prev=node->prev;
		free(node);
	}
	return(root);
}

int setcurr(root, obj)
list root;
void *obj;
{
	list node=root->obj;

	if(node!=root)
	{
		node->obj=obj;
		return(1);
	}
	else
		return(0);
}

void freelist(root)
/* frees the memory for the list (but not the objects) */
list root;
{
	list tmp;

	while(root!=NULL)
	{
		tmp=root->next;
		free(root);
		root=tmp;
	}
}

list listnode(root)
list root;
/* returns a pointer to the current node */
{
	return(root->obj);
}

list setnode(root, node)
list root, node;
/* sets the current node to be node */
/* returns root if ok, 0 if not */
{
	root->obj=node;
	return(root);
}

list endlist(root)
/* set current to last node in list */
list root;
{
	root->obj=root;
	while(root->obj->next!=NULL)
		root->obj=root->obj->next;
	return(root);
}

list startlist(root)
/* sets current to before first in list */
/* (next will give first node) */
list root;
{
	root->obj=root;
	return(root);
}

list listgonext(root)
/* sets current to next node in list */
list root;
{
	if(root->obj->next!=NULL)
		root->obj=root->obj->next;
	return(root);
}

list lastnode(root)
/* returns last node in list */
list root;
{
	if(root!=NULL)
		while(root->next!=NULL)
			root=root->next;
	return(root);
}

list listgoprev(root)
/* sets current to the previous node in list */
list root;
{
	if(root->obj->prev!=root)
		root->obj=root->obj->prev;
	return(root);
}

void *listnext(root)
/* sets current to next node in list, and returns that object */
list root;
{
	if(root->obj->next==NULL)
		return(NULL);
	else
		return((root->obj=root->obj->next)->obj);
}

void *listprev(root)
/* sets current to the previous node in list, and returns that object */
list root;
{
	if(root->obj->prev==root)
		return(NULL);
	else
		return((root->obj=root->obj->prev)->obj);
}

void *listcurr(root)
/* returns the current object */
list root;
{
	if(root->obj==root)
		return(NULL);
	else
		return(root->obj->obj);
}

int listlastp(root)
/* returns true if current is last in list */
list root;
{
	return(root->obj->next==NULL);
}

int listfirstp(root)
/* returns true if current is first in list */
list root;
{
	return(root->obj->prev==root);
}

int listemptyp(root)
/* returns true if there is nothing in the list */
list root;
{
	return(root->next==NULL);
}

#define firstnode(l) ((l)==NULL?NULL:((list)(l))->next)
#define nextnode(l) ((l)==NULL?NULL:((list)(l))->next)
#define prevnode(l) ((l)==NULL?NULL:(((list)(l))->prev==NULL?NULL:           (((list)((list)(l))->prev)->prev==NULL?NULL:((list)(l))->prev)))
#define lfor(l,n) for(n=firstnode(l);n!=NULL;n=nextnode(n))

#define nodeobj(n) (((list)n)->obj)
#define setobj(n, i) (((list)n)->obj=(list)(i))

#define MAXLEN 256
#define compare(s, t) (strncmp(s, t, strlen(t))==0)
/* number of groups of in genbank sequence,
 * number of bases per group,
 * calculation macro for position of group */
#define NUMGROUPS 6
#define GROUPLEN 10
#define group_pos(i) (10+(i)*(GROUPLEN+1))
/* format for number on genbank sequence line */
#define NUMFORMAT "%9ld"
/* calculation of whether character is a spacing character */
#define space_char(c) (!((c)=='-'||(c)=='.'||(c)=='~'))

#define FORMAT_NUM_BASE_PAIR "%6d"

/* Token strings */
#define ENDENTRY "//"
#define LOCUSLINE "LOCUS"
#define BASECOUNTLINE "BASE COUNT"
#define ORIGINLINE "ORIGIN"

/* Error strings */
/* for getnames() */
#define INVALID_RANGE "Invalid range specified with '%s'\n"
#define IGNORE_RANGE "Ignoring range specified with '%s'\n"
#define ALREADY_ONE "\t(there's already one specified with '%s')\n"
#define START_TOO_LOW "\t(it's starting base is less than 1)\n"
#define END_BEFORE_START "\t(it's ending base is before it's starting base)\n"
#define PROBLEMS "Names read successfully, although problems (noted above) were encountered.\n"
#define SUCCESSFUL "Names read successfully.\n"
/* for getsequence */
#define END_OF_FILE_INTERRUPTS "End_of_file interupts sequence\n"
#define END_OF_ENTRY_INTERRUPTS "End_of_data interupts sequence\n"
#define IRREGULAR_SEQUENCE_NUMBERING "Irregular sequence numbering\n"
#define NAME_NOT_IN_GENBANK "Organism '%s' not found in alignment\n"
#define NO_NAMES_IN_GENBANK "No Organisms were found in the alignment\n"
#define SEQUENCE_LENGTH_SMALL "Sequence length is less than the expected number of Base Pairs\n"
#define SEQUENCE_LENGTH_BIG "Sequence length is more than the expected number of Base Pairs.  Truncating...\n"

typedef struct {
	char *name;			/* name specified in subalignment file */
	int base_pairs;		/* number of output base pairs */
	list genbank_lines;	/* lines of genbank entry */
	int len;			/* length of sequence */
	char *sequence;		/* sequence characters */
	int data_line;		/* 1 if there is no sequence data in this
						   entry */
	} organismd, *organism;

#define thisint(t) ((int)nodeobj(t))
#define thisorg(t) ((organism)nodeobj(t))

char *dupstring(s)
char *s;
{
	char *tmp;

	tmp=(char *)malloc(strlen(s)+1);
	strcpy(tmp, s);
	return(tmp);
}

/* calculation of whether character should be counted as a base */

int base(c)
char c;
{
    switch(c)
    {
	case 'T':
	case 'U':
	case 'G':
	case 'K':
	case 'C':
	case 'Y':
	case 'S':
	case 'B':
	case 'A':
	case 'W':
	case 'R':
	case 'D':
	case 'M':
	case 'H':
	case 'V':
	case 'N':
	case 't':
	case 'u':
	case 'g':
	case 'k':
	case 'c':
	case 'y':
	case 's':
	case 'b':
	case 'a':
	case 'w':
	case 'r':
	case 'd':
	case 'm':
	case 'h':
	case 'v':
	case 'n':
	    return(1);
	default:
	    return(0);
    }
}

printseq(seq, len)
char *seq;
int len;
{
	int i;

	for(i=0;i<len;i++)
		printf("%c", seq[i]);
	printf("\n");
}
printmask(mask, len)
char *mask;
int len;
{
	int i;

	for(i=0;i<len;i++)
		if(mask[i])
			printf("1");
		else
			printf("0");
	printf("\n");
}

char *getlin(f, endchar)
/* get a line and return it */
FILE *f;
int *endchar;
{
	char buf[MAXLEN+1];
	int i=0;
	int ch;

	ch=fgetc(f);
	if(ch==EOF)
	{
		*endchar=ch;
		return(NULL);
	}
	while(i<MAXLEN)
	{
		if(ch=='\n' || ch==EOF)
			break;
		else
			buf[i++]=ch;
		ch=fgetc(f);
	}
	if(i==MAXLEN)
	/* not enough room to store line */
	{
		/* read rest of line */
		while(ch!=EOF && ch!='\n')
			ch=fgetc(f);
	}
	*endchar=ch;
	buf[i]='\0';
	return(dupstring(buf));
}

organism find_org(orgs, name)
/* sets current in orgs to be the organism with the given name */
/* returns the organism, or NULL if not found */
list orgs;
char *name;
{
	organism taxa;

	/* find the name in the list */
	startlist(orgs);
	while((taxa=listnext(orgs))!=NULL)
		if(strcmp(taxa->name, name)==0)
			break;

	/* if the name wasn't found, set so that nothing is current */
	if(taxa==NULL)
	{
		startlist(orgs);
		return(NULL);
	}
	else
	{
		return(taxa);
	}
}

list getnames(in, log)
/* reads the names from in, and makes a list out of them */
/* sets current element in list to be the one with the range on it */
/* logs errors in log */
FILE *in, *log;  //++pac
{
    char *l;
    list orgs=newlist();
    char *index;
    organism tmp;
    char name[MAXLEN+1];
    int i;
    int problems=0;
    int ch;

    if(orgs==NULL)
		return(NULL);
    while((l=getlin(in, &ch))!=NULL)
    {
		i=sscanf(l, "%s", name);
		if(i==1)
		{
			/* name successfully read */
			tmp=(organism)malloc(sizeof(organismd));
			tmp->name=dupstring(name);
			tmp->sequence=NULL;
			tmp->genbank_lines=NULL;
			if(!addnode(orgs, tmp))
			{
				/* not enough memory to create list */
				return(NULL);
			}
		}
		free(l);
    }
    if(problems)
		fprintf(log, PROBLEMS);
    else
		fprintf(log, SUCCESSFUL);
    /* set current to be range_org, the one with the range */
	startlist(orgs);
    return(orgs);
}

int read_sequence(gen, seq, data_line, last_line, log)
FILE *gen;
char *seq;
int *data_line;
char **last_line;
FILE *log;
{
	char buf[80];		/* holds characters of block */
	char *l;		/* current line */
	int ll;		/* length of current line */
	int length;		/* length of sequence read */
	int i,j,k;
	int ch;
	int group;		/* group number we are on */
	int index;		/* true index at start of line */
	int irregular;		/* flag if numbering is irregular */
	int eof_in_sequence;	/* flag if EOF appears inside the sequence */

	length=0;
	irregular=0;
	eof_in_sequence=0;

	l=getlin(gen, &ch);
	ll=strlen(l);
	do
	{
		/* skip space, looking for number */
	    for(i=0;isspace(l[i]);i++);
	    if(!isdigit(l[i]))
			/* if first character is not 1-9, done with sequence */
			break;
	    i=sscanf(l, "%ld", &index);
	    if(i!=1)
			/* no number at beginning of line, done with sequence */
			/* realistically, this should never happen, since we have
				checked for a number above */
			break;
	    if(index!=length+1 && !irregular)
	    {
			/* bad numbering */
			fprintf(log,IRREGULAR_SEQUENCE_NUMBERING);
			/* flag, so we don't print this message again.  We will still
				attempt to read the sequence */
			irregular=1;
	    }
		i=0;
		/* find number */
		while(isspace(l[i]))
			i++;
		/* skip number */
		while(isdigit(l[i]))
			i++;
		j=0;
		while(l[i]!='\0' && l[i]!=EOF)
		{
			if(isspace(l[i]))
				i++;
			else
			{
				if(base(l[i]))
					*data_line=0;
				buf[j++]=l[i++];
			}
		}
		buf[j]='\0';
		add_string(seq, buf);
		length+=j;
	    /* now have added all sequence that exists on this line to seq */
	    if(l[i]==EOF)
	    {
			/* must stop, since at end of file */
			eof_in_sequence=1;
			break;
	    }
	    else
	    {
			l=getlin(gen, &ch);
			ll=strlen(l);
	    }
	} while(1);
	*last_line=l;
	if(eof_in_sequence)
	{
	    fprintf(log,END_OF_FILE_INTERRUPTS);
	    *last_line=0;
	}
	return(length);
}

char *getsequence(gen, log, len, last_line, data_line)
/* read a sequence from gen (genbank) */
/* return the last line read (the terminating line) in last_line */
/* write errors to log */
/* sets data_line to 0 if sequence has bases in it */
/* returns length in len */
FILE *gen, *log;
int *len;
char **last_line;
int *data_line;
{
	char *seq;	/* alignment string */

	/* create a new dynamic string */
	seq=new_string();
	*len=read_sequence(gen, seq, data_line, last_line, log);
	return(freeze_string(seq));
}

free_line_list(l)
/* frees the data for the line list */
{
	char *line;

	/* first free each line */
	startlist(l);
	while((line=listnext(l))!=NULL)
		free(line);

	/* now free list structure */
	freelist(l);
}

getentries(orgs, gen, log)
/* store the entries from gen (genbank) in the orgs list */
/* write log info to log */
list orgs;
FILE *gen, *log;
{
    char *l;
    list lines;	/* lines for the current genbank entry */
    int requested_entry;	/* =1 if it's still possible that this entry */
				/*    was requested */
    organism taxa;		/* organism that the entry matches */
    char *sequence;		/* temporary space for reading in sequence */
    char name[MAXLEN];
	int len;
    int i;
    int ch;
    int data_line;
    int dummy;
	int more;

    while((l=getlin(gen, &ch))!=NULL)
    {
		/* have first line in entry */
		lines=newlist();
		requested_entry=1;
		taxa=NULL;
		sequence=NULL;
		more=1;

		while(more)
		{
			if(l==NULL)
			{
				fprintf(stderr,
					"ERROR--Unexpected EOF in genbank file\n");
				exit(0);
			}
			else if(compare(l, ENDENTRY))
				more=0;
			else if(requested_entry)
			{
				if(!addnode(lines, l))
				{
					/* not enough memory to create add line to entry */
					fprintf(stderr,
						"ERROR--Not enough memory to read genbank file\n");
					exit(0);
				}
				if(compare(l, LOCUSLINE))
				{
					/* find if name is in orgs */
					i=sscanf(l, "LOCUS %s BP", name);
					if(i==0 || i==EOF)
					{
						/* badly formatted LOCUS line */
						/* so we will ignore this entry */
						requested_entry=0;
					}
					else
					{
						if((taxa=find_org(orgs, name))!=NULL)
							/* organism was requested */
							/* now taxa=the organism for this entry */
							;
						else
						{
							/* organism was not requested */
							requested_entry=0;
						}
					}
					/* get line after locus line */
					l=getlin(gen, &ch);
				}
				else if(compare(l, ORIGINLINE))
				{
					if(sequence==NULL)
					{
						/* read the sequence after the origin line */
						data_line=1;
						sequence=getsequence(gen, log, &len, &l, &data_line);
						/* if l is returned, it's the line in gen, so no need
							to getlin() */
						if(l==NULL)
							l=getlin(gen, &ch);
						/* store length */
						taxa->len=len;
					}
					else
					{
						/* more than one sequence per this entry, so ignore */
						char *tmp=getsequence(gen, log, &len, &l, &dummy);
						if(tmp!=NULL)
							free(tmp);
						/* if l is returned, it's the line in gen, so no need
							to getlin() */
						if(l==NULL)
							l=getlin(gen, &ch);
					}
				}
				else
					/* get next line */
					l=getlin(gen, &ch);
			}
			else
				/* get next line if not requested_entry */
				l=getlin(gen, &ch);
		}
		if(requested_entry && taxa!=NULL)
		{
			/* setup taxa */
			taxa->genbank_lines=lines;
			taxa->sequence=sequence;
			taxa->data_line=data_line;
		}
		else
		{
			/* throw away lines that were read in */
			free_line_list(lines);
			/* throw away the sequence if it was read */
			if(sequence!=NULL)
				free(sequence);
		}
    }
    /* now we've read in all the entries */
    /* so we can return */
}

int get_maximum_length(orgs)
list orgs;
{
	int maxl;
	list t=listnode(orgs);
	organism o;

	maxl=0;
	startlist(orgs);
	while((o=(organism)listnext(orgs))!=NULL)
		if(o->len>maxl)
			maxl=o->len;

	setnode(orgs, t);
	return(maxl);
}

int calc_num_base_pairs(orgs, mask, mask_len)
/* counts non-zeros in mask, and calculates number of sequence characters for
	each organism */
char *mask;
int mask_len;
{
	int i, num;
	list t=listnode(orgs);
	organism o;

	startlist(orgs);
	while((o=(organism)listnext(orgs))!=NULL)
	{
		num=0;
		for(i=0;i<o->len;i++)
			if(mask[i] && base(o->sequence[i]))
				num++;
		o->base_pairs=num;
	}

	setnode(orgs, t);
}

void countbases(seq, mask, mask_len, a, c, g, t, others)
char *seq;
char *mask;
int mask_len;
int *a, *c, *g, *t, *others;
{
	int i;

	*a=0;
	*c=0;
	*g=0;
	*t=0;
	*others=0;
	for(i=0;i<mask_len;i++)
		if(mask[i])
		{
			switch(seq[i])
			{
				case 'a':
				case 'A':
					(*a)++;
					break;
				case 'c':
				case 'C':
					(*c)++;
					break;
				case 'g':
				case 'G':
					(*g)++;
					break;
				case 't':
				case 'T':
				case 'u':
				case 'U':
					(*t)++;
					break;
				default:
					(*others)++;
					break;
			}
		}
}

printentry(taxa, mask, mask_len, out)
/* print the entry for taxa (masked by mask) into file out */
organism taxa;
char *mask;
int mask_len;
FILE *out;
{
    char *l;
    int i, j;

    startlist(taxa->genbank_lines);
    while((l=listnext(taxa->genbank_lines))!=NULL)
    {
		/* print out each line */
		if(compare(l, LOCUSLINE))
		{
			/* put updated num base pair on LOCUS line */
			/* first find old number */
			for(i=0;!isspace(l[i]);i++); /* skip LOCUS */
			for(;isspace(l[i]);i++);	 /* find name */
			for(;!isspace(l[i]);i++); /* skip name */
			for(;!isdigit(l[i]);i++);	 /* find basepair number */
			for(;isdigit(l[i]);i++); /* find end of number */

			sprintf(&l[i-6], FORMAT_NUM_BASE_PAIR, taxa->base_pairs);
			/* get rid of '\0' that sprintf adds after number */
			l[strlen(l)]=' ';
		}
		else if(compare(l, BASECOUNTLINE))
		{
			int a, c, g, t, others;
			char buf[80];
			countbases(taxa->sequence, mask, taxa->len,
				&a, &c, &g, &t, &others);
			if(others==0)
				sprintf(buf, "BASE COUNT  %6d a %6d c %6d g %6d t",
					a, c, g, t);
			else
				sprintf(buf, "BASE COUNT  %6d a %6d c %6d g %6d t %6d others",
					a, c, g, t, others);

			l=dupstring(buf);
		}
		fprintf(out, "%s\n", l);
		if(compare(l, ORIGINLINE))
		{
			/* just printed out ORIGIN line, so now print out sequence */
			int pos, grouppos, groupnum;
			int state;
#define FirstOnLine		0
#define FirstInGroup	2
#define InsideGroup		1
#define NoMoreBases		3

			pos=0;
			i= -1;
			state=FirstOnLine;
			while(state<NoMoreBases)
			{
				/* go to next outputtable character */
				for(i++;i<taxa->len;i++)
					if(mask[i])
						break;
				pos++;
				switch(state)
				{
					case FirstOnLine:
						grouppos=1;
						groupnum=0;
						if(i>=taxa->len)
							state=NoMoreBases;
						else
						{
							fprintf(out, NUMFORMAT, pos);
							fprintf(out, " %c", taxa->sequence[i]);
							state=InsideGroup;
						}
						break;
					case FirstInGroup:
						grouppos=1;
						if(i>=taxa->len)
						{
							fprintf(out, "\n");
							state=NoMoreBases;
						}
						else
						{
							fprintf(out, " %c", taxa->sequence[i]);
							state=InsideGroup;
						}
						break;
					case InsideGroup:
						if(i>=taxa->len)
						{
							fprintf(out, "\n");
							state=NoMoreBases;
						}
						else
						{
							fprintf(out, "%c", taxa->sequence[i]);
							grouppos++;
							if(grouppos==GROUPLEN)
							{
								groupnum++;
								if(groupnum==NUMGROUPS)
								{
									state=FirstOnLine;
									fprintf(out, "\n");
								}
								else
									state=FirstInGroup;
							}
						}
						break;
					case NoMoreBases:
						break;
				}
			}
		}
    }
    fprintf(out, "%s\n", ENDENTRY);
}

int prevchar(i, mask, len)
/* return the previous position with ch in mask */
/* returns -1 if none */
int i;
char *mask;
int len;
{
	for(i--;i>-1;i--)
	{
		if(mask[i]==1)
			return(i);
		if(mask[i]==2)
			return(-1);
	}
	return(-1);
}

int nextchar(i, mask, len)
/* return the next position with ch in mask */
/* returns -1 if none */
int i;
char *mask;
int len;
{
	for(i++;i<len;i++)
	{
		if(mask[i]==1)
			return(i);
		if(mask[i]==2)
			return(-1);
	}
	return(-1);
}

removegaps(orgs, mask, start, end, mask_len)
/* remove common gaps in the orgs alignment, by placing 0's in mask */
/* ignore columns already masked out in mask */
list orgs;
char *mask;
int start, end;
int mask_len;
{
    int i;		/* current column */
    int a, b;
    organism taxa;
    int deletable;	/* =1 if you can delete current column */
    int have_deleted;	/* 1 if we have deleted a column on this pass */
    int must_check_right;	/* 1 if we still must check char on right */

    do
    {
		have_deleted=0;
		for(i=start;i<end;i++)
		{
			if(mask[i]!=0)
			{
				startlist(orgs);
				deletable=1;
				while((taxa=listnext(orgs))!=NULL && deletable)
				{
					if(i>taxa->len)
						deletable=0;
					else
						switch(taxa->sequence[i])
						{
							case '-':
								break;
							case '.':
								must_check_right=1;
								a=prevchar(i, mask, mask_len);
								if(a!=-1)
									if(taxa->sequence[a]=='.')
										must_check_right=0;
								if(must_check_right)
								{
									b=nextchar(i, mask, mask_len);
									if(b==-1)
										deletable=0;
									else if(taxa->sequence[b]!='.')
										deletable=0;
								}
								break;
							case '~':
								must_check_right=1;
								a=prevchar(i, mask, mask_len);
								if(a!=-1)
									if(taxa->sequence[a]=='~')
										must_check_right=0;
								if(must_check_right)
								{
									b=nextchar(i, mask, mask_len);
									if(b==-1)
										deletable=0;
									else if(taxa->sequence[b]!='~')
										deletable=0;
								}
								break;
							default:
								/* it's a base or metacharacter */
								deletable=0;
								break;
						}
				}
				if(deletable)
				{
					/* then didn't find any with a base here, so deletable */
					mask[i]=0;
					have_deleted=1;
				}
			}
		}
    }while(have_deleted);
}

int read_pair(s, start, end, astart, aend)
char *s;
int *start, *end;
int astart, aend;
/* returns 0--ok, 1--no ',', 2--error reading number */
{
	int i;

	if(*s=='\0')
		return(1);
	else if(*s==',')
	{
		*start=astart;
		s++;
	}
	else
	{
		i=sscanf(s, "%d", start);
		if(i==0)
			return(2);
		while(*s!=',' && *s!='\0')
			s++;
		if(*s!=',')
			return(1);
		s++;
	}
	if(*s=='\0')
		*end=aend;
	else
	{
		i=sscanf(s, "%d", end);
		if(i==0)
			return(2);
	}
	return(0);
}

int read_pair_list(s, starts, ends, astart, aend, errorstring)
char *s;
list starts, ends;
int astart, aend;
char *errorstring;
{
	int i, j, ss, ee, os, oe, num, error;

	startlist(starts);
	startlist(ends);

	os= -1;
	oe= -1;
	num=1;

	for(;*s!='\0';)
	{
		if(num!=1)
			s++;
		error=read_pair(s, &ss, &ee, astart, aend);
		if(error)
		{
			fprintf(stderr, "ERROR--%s: Error reading Range %d\n",
				errorstring, num);
			return(1);
		}

		if(ee<ss)
		{
			fprintf(stderr, "ERROR--%s: End before Start in Range %d\n",
				errorstring, num);
			return(1);
		}
		else if(ss<=0)
		{
			fprintf(stderr, "ERROR--%s: Range %d starts at less than 1\n",	
				errorstring, num);
			return(1);
		}
		else if(ss<os)
		{
			fprintf(stderr, "ERROR--%s: Range %d covers other ranges\n",
				errorstring, num);
			return(1);
		}
		else if(ss<=oe)
		{
			fprintf(stderr,
				"ERROR--%s: Ranges %d and %d overlap\n",
				errorstring, num-1, num);
			return(1);
		}
		else
		{
			addnode(starts, ss);
			addnode(ends, ee);
		}
		os=ss;
		oe=ee;
		num++;
		while(*s!=':' && *s!='\0')
			s++;
	}
	return(0);
}

#ifdef DEBUG
psl(s, e)
list s, e;
{
	int ss, ee;

	startlist(s);
	startlist(e);
	while((ss=(int)listnext(s))!=NULL)
	{
		ee=(int)listnext(e);
		printf("%d,%d:", ss, ee);
	}
	printf("\n");
}
#endif

int calc_seq_cols(seq_starts, seq_ends, seq_col_starts, seq_col_ends,
	reference)
/* updates seq_col_starts,ends */
/* returns 0 if error found in sequence positions */
list seq_starts, seq_ends, seq_col_starts, seq_col_ends;
organism reference;
{
	char *s;
	list start, end;
	int ss, ee;
	int curbase, curcol;
	int first;
	int need_end;
	int range_num;

	curbase=0;
	curcol= -1;
	start=firstnode(seq_starts);
	end=firstnode(seq_ends);
	startlist(seq_col_starts);
	startlist(seq_col_ends);
	need_end=0;
	range_num=1;
	if(start!=NULL && end!=NULL)
	{
		for(s=reference->sequence;*s!='\0';s++)
		{
			curcol++;
			if(base(*s))
				curbase++;
			if(start!=NULL && curbase==thisint(start))
			{
				addnode(seq_col_starts, curcol);
				need_end=1;
				start=nextnode(start);
			}
			if(end!=NULL && curbase==thisint(end))
			{
				addnode(seq_col_ends, curcol);
				need_end=0;
				end=nextnode(end);
				range_num++;
			}
		}
		if(need_end)
		{
			fprintf(stderr,
				"ERROR--Range %d extends past end of reference sequence\n",
				range_num);
			return(0);
		}
		else if(start!=NULL)
		{
			fprintf(stderr,
				"ERROR--Range %d starts past end of reference sequence\n",
				range_num);
			return(0);
		}
	}
	return(1);
}

int reconfigure_sequences(starts, ends, newlen, orgs)
list starts, ends;
int newlen;
list orgs;
{
	list o;
	char *ts;
	char *ns;
	list start, end;
	int i, j, first;

	lfor(orgs, o)
	{
		ts=thisorg(o)->sequence;
		ns=(char *)malloc(newlen+1);
		if(ns==NULL)
			return(0);
		first=1;
		i=0;
		for(start=firstnode(starts),end=firstnode(ends);
				start!=NULL;
				start=nextnode(start),end=nextnode(end))
		{
			if(first)
				first=0;
			else
				ns[i++]='|';
			if(thisint(start)!=0)
			{
				/* fill in prev column */
				if(thisint(start)>=thisorg(o)->len)
					ns[i++]='.';
				else if(base(ts[thisint(start)]))
				{
					if(base(ts[thisint(start)-1]))
						ns[i++]='-';
					else
						ns[i++]=ts[thisint(start)-1];
				}
				else
					ns[i++]=ts[thisint(start)];
			}
			/* now i = where to start storing this range */
			for(j=thisint(start);j<=thisint(end);j++)
				if(j>=thisorg(o)->len)
					ns[i++]='.';
				else
					ns[i++]=ts[j];
			if(thisint(end)!=newlen-1)
			{
				/* fill in following column */
				if(thisint(end)>=thisorg(o)->len)
					ns[i++]='.';
				else if(base(ts[thisint(end)]))
				{
					if(base(ts[thisint(end)+1]))
						ns[i++]='-';
					else
						ns[i++]=ts[thisint(end)+1];
				}
				else
					ns[i++]=ts[thisint(end)];
			}
		}
		ns[i]='\0';
		free(thisorg(o)->sequence);
		thisorg(o)->sequence=ns;
		thisorg(o)->len=i;
	}
	return(1);
}

int reconfigure_alignment(seq_starts, seq_ends, seq_col_starts, seq_col_ends,
	col_start, col_end,
	seq_specified, col_specified, orgs, reference, alignment_length, mask)
list seq_starts, seq_ends;
list seq_col_starts, seq_col_ends;
int col_start, col_end;
int col_specified, seq_specified;
list orgs;
organism reference;
int *alignment_length;
char **mask;
{
	list start, end;
	int newlen, first;
	int i, di;
	char *tm;

	col_start--;
	col_end--;

	/* reconfigure alignment for subranges */

	if(!col_specified && !seq_specified)
	{
		startlist(seq_col_starts);
		startlist(seq_col_ends);
		addnode(seq_col_starts, 0);
		addnode(seq_col_ends, (*alignment_length)-1);
		*mask=(char *)malloc(*alignment_length);
		if(*mask==NULL)
		{
			/* not enough memory to create mask */
			fprintf(stderr, "ERROR--Not enough memory to store sequences\n");
			return(0);
		}

		tm= *mask;

		/* reset mask to include all columns */
		for(i=0;i<*alignment_length;i++)
			tm[i]=1;
		return(1);
	}


	if(!seq_specified)
	{
		/* deal with columns only */
		startlist(seq_col_starts);
		startlist(seq_col_ends);
		addnode(seq_col_starts, col_start);
		addnode(seq_col_ends, col_end);
	}
	else
	{
		if(!col_specified)
		{
			col_start=0;
			col_end=(*alignment_length)-1;
		}
		if(!calc_seq_cols(seq_starts, seq_ends, seq_col_starts, seq_col_ends,
				reference))
			/* error in seq positions */
			return(0);
		/* prune to stuff specified by column */
		for(start=firstnode(seq_col_starts),end=firstnode(seq_col_ends);
				start!=NULL;
				start=nextnode(start),end=nextnode(end))
		{
			if(thisint(end)<col_start)
			{
				/* remove this subrange */
				start=rmnode(start);
				end=rmnode(end);
			}
			else if(thisint(start)<col_start)
				setobj(start, col_start);
			else if(thisint(start)>col_end)
			{
				/* remove this subrange */
				start=rmnode(start);
				end=rmnode(end);
			}
			else if(thisint(end)>col_end)
				setobj(end, col_end);
		}
	}

	/* calculate new alignment length */
	newlen=0;
	first=1;
	for(start=firstnode(seq_col_starts),end=firstnode(seq_col_ends);
			start!=NULL;
			start=nextnode(start),end=nextnode(end))
	{
		if(first!=1)
			/* column for '|' seperator */
			newlen++;
		first=0;
		if(thisint(start)!=1)
			/* previous column */
			newlen++;
		if(thisint(end)!=*alignment_length)
			/* following column */
			newlen++;
		newlen+=thisint(end)-thisint(start)+1;
	}

	/* for each sequence, copy the relevant stuff into a new sequence buf */
	if(!reconfigure_sequences(seq_col_starts, seq_col_ends, newlen, orgs))
	{
		fprintf(stderr, "ERROR--Not enough memory to store sequences\n");
		return(0);
	}

	/* recalc seq_cols */
	first=1;
	for(start=firstnode(seq_col_starts),end=firstnode(seq_col_ends);
			start!=NULL;
			start=nextnode(start),end=nextnode(end))
	{
		if(first)
		{
			if(thisint(start)==0)
				i=0;
			else
				i=1;
			first=0;
		}
		/* get length of subrange */
		di=thisint(end)-thisint(start)+1;
		/* translate range to begin at i (3 past last subrange) */
		setobj(start, i);
		setobj(end, i+di);
		/* calculate start of next subrange (skip 2 spacing cols and '|')*/
		i+=di+3;
	}


	/* alloc mask, set according to seq_cols */
	*mask=(char *)malloc(newlen);
	if(*mask==NULL)
	{
		/* not enough memory to create mask */
		fprintf(stderr, "ERROR--Not enough memory to store sequences\n");
		return(0);
	}

	tm= *mask;
	for(i=0;i<newlen;i++)
		tm[i]=2;
	for(start=firstnode(seq_col_starts),end=firstnode(seq_col_ends);
			start!=NULL;
			start=nextnode(start),end=nextnode(end))
	{
		for(i=thisint(start);i<=thisint(end);i++)
			tm[i]=1;
		if(thisint(start)>0)
			tm[thisint(start)-1]=1;
		if(thisint(end)<newlen)
			tm[thisint(end)+1]=1;
	}

	/* store newlen in alignment_length */
	*alignment_length=newlen;
}

printhelp(longhelp, commandname)
int longhelp;
char *commandname;
{
	if(longhelp)
	{
fprintf(stderr, "\
usage: %s input-file alignment-file output-file\n\
  options:\n\
      -pack|nopack\n\
           -specifies packing option, defaults to '-pack'\n\
           -pack specifies to remove common gaps from the output alignment\n\
           -nopack specifies turns off packing\n\
\n\
      -reference=<organism name>\n\
           -specifies reference organism for '-positions' option\n\
           -defaults to E.coli, or first organism if not specified\n\
           -abbreviate as '-r=<organism name>'\n\
\n\
           -Example:  -r=E.coli\n\
\n\
      -columns=[start],[end]\n\
           -specifies a section of the alignment by alignment column\n\
           -[start] defaults to beginning of alignment\n\
           -[end] defaults to end of alignment\n\
           -abbreviate as -c=[start],[end]\n\
\n\
           -Example:  -c=540,623\n\
\n\
      -positions=[start],[end][:[start],[end]]...\n\
           -specifies sections of the alignment by sequence position\n\
           -may specify list of non-overlapping ranges\n\
                overlapping or out of order ranges are not valid, contiguous\n\
                ranges will be accepted but not consolidated (contiguous\n\
                positions may not be in contiguous columns...)\n\
           -in output alignment, sections are seperated by vertical bar\n\
                character: '|'\n\
           -sequence positions refer to position in reference organism,\n\
                see '-reference' option above\n\
           -[start] defaults to beginning of sequence\n\
           -[end] defaults to end of sequence\n\
           -abbreviate as -p=[start],[end][:[start],[end]]...\n\
\n\
           -Example:  -p=15,30:110,125\n\
           -Another:  -p=,30:40,50:60,\n\
           \n\
", commandname);
	}
	else
	{
fprintf(stderr, "\
usage: %s input-file alignment-file output-file\n\
  options:\n\
      -pack|nopack\n\
      -reference=<organism name>\n\
      -columns=[start],[end]\n\
      -positions=[start],[end][:[start],[end]]...\n\
type 'subalign -help' for more info\n\
", commandname, commandname);
	}
}


main(argc, argv)
int argc;
char **argv;
{
    FILE *in, *out, *log, *genbank; /* files to be used */
    char *l;
    list orgs;				/* list of organisms */
    organism taxa;			/* temporary organism pointer */
    organism reference;		/* reference organism */
	int reference_length;
    int alignment_length;	/* length of each alignment sequence */
    int new_length;	/* length of each alignment sequence */
    char *mask;				/* for storing which columns to use */
    int i;
    list previous_organism;
    int ch;
	int packing;
	int innum, genbanknum, outnum, toomany;
	int column_specified, sequence_specified, reference_specified;
	int column_error, sequence_error, reference_error;
	char *reference_name;
	int more;
	list seq_starts, seq_ends;
	int col_start, col_end;
	list seq_col_starts, seq_col_ends;
	int error;


	seq_starts=newlist();
	seq_ends=newlist();
	seq_col_starts=newlist();
	seq_col_ends=newlist();


	packing=1;
	reference_specified=0;
	reference_error=0;
	innum=outnum=genbanknum=toomany=0;
	more=1;
	error=0;
	for(i=1;i<argc && more;i++)
	{
		if(argv[i][0]=='-')
		{
			
			if(compare(argv[i], "-help"))
			{
				printhelp(1, argv[0]);
				exit(0);
			}
			else if(compare(argv[i], "-pack"))
				packing=1;
			else if(compare(argv[i], "-nopack"))
				packing=0;
			else if(argv[i][1]=='c')
			{
				/* columns */
				/* deal with later */
			}
			else if(argv[i][1]=='p')
			{
				/*position*/
				/* deal with later */
			}
			else if(argv[i][1]=='r')
			{
				/*reference*/

				static int len, j;
				static int error;

				len=strlen(argv[i]);

				for(j=0;j<len;j++)
					if(argv[i][j]=='=')
						break;
				if(j==len)
				{
					fprintf(stderr, "ERROR--No name specified on -r option\n");
					printhelp(0, argv[0]);
					exit(1);
				}
				else if(argv[i][j+1]=='\0')
				{
					fprintf(stderr,
						"ERROR--Reference name on -r option is empty\n");
					printhelp(0, argv[0]);
					exit(1);
				}
				else
				{
					reference_name=dupstring(&argv[i][j+1]);
					reference_specified=1;
				}
			}
			else
			{
				fprintf(stderr,
					"ERROR--Invalid option: %s\n", argv[i]);
				printhelp(0, argv[0]);
				exit(1);
			}
				
		}
		else
		{
			if(innum==0)
				innum=i;
			else if(genbanknum==0)
				genbanknum=i;
			else if(outnum==0)
				outnum=i;
			else
				toomany=1;
		}
	}

	if(toomany || outnum==0)
	{
		if(toomany)
			fprintf(stderr,
				"ERROR--Too Many file names\n");
		else
			fprintf(stderr,
				"ERROR--Not enough file names\n");
		printhelp(0, argv[0]);
		exit(1);
	}
		
    in=fopen(argv[innum], "r");
    genbank=fopen(argv[genbanknum], "r");
    out=fopen(argv[outnum], "w");
    log=stderr;
    if(in==NULL ||
	    out==NULL ||
	    log==NULL ||
	    genbank==NULL)
    {
	    fprintf(stderr,
			"ERROR--Could not open files:");
	    if(in!=NULL)
		    fclose(in);
	    else
		    fprintf(stderr, "\n\tinput ('%s')", argv[innum]);
	    if(out!=NULL)
		fclose(out);
	    else
		    fprintf(stderr, "\n\toutput ('%s')", argv[outnum]);
	    if(log!=NULL)
	    {
		    if(log!=stderr)
			fclose(log);
	    }
	    else
		    fprintf(stderr, "\n\tlog ('%s')");
	    if(genbank!=NULL)
		    fclose(genbank);
	    else
		    fprintf(stderr, "\n\talignment ('%s')", argv[genbanknum]);
	    fprintf(stderr, ".\n");
	    exit(1);
    }

    /* in file ok, genbank file ok, so start reading names */
    orgs=getnames(in, log);

    /* names read, so start reading their sequence entries */
    getentries(orgs, genbank, log);

    /* print error messages for names not found in genbank */
    startlist(orgs);
    taxa=listnext(orgs);
    while(taxa!=NULL)
	    if(taxa->genbank_lines==NULL)
	    {
		    /* name not in alignment, so remove it */
		    if(taxa==reference)
			    reference=NULL;
		    fprintf(log,NAME_NOT_IN_GENBANK,taxa->name);
		    free(taxa->name);
		    if(listlastp(orgs))
		    {
			    rmcurr(orgs);
			    taxa=NULL;
		    }
		    else
		    {
			    rmcurr(orgs);
			    taxa=listcurr(orgs);
		    }
	    }
	    else
		    taxa=listnext(orgs);

    startlist(orgs);
    if(listlastp(orgs))
    {
	    /* no names found in alignment */
	    fprintf(log,NO_NAMES_IN_GENBANK);
	    fclose(in);
	    fclose(out);
	    if(log!=stderr)
		fclose(log);
	    fclose(genbank);
	    exit(0);
    }

    /* get reference organism with the range on it */
	{
		list t;
		organism ecoli;
		organism first=NULL;
		char *s;

		reference=NULL;

		lfor(orgs, t)
		{
			if(first==NULL)
				first=thisorg(t);
			if(strcmp(thisorg(t)->name, "E.coli")==0)
				ecoli=thisorg(t);
			if(reference_specified)
			{
				if(strcmp(thisorg(t)->name, reference_name)==0)
				{
					reference=thisorg(t);
					break;
				}
			}
		}

		if(reference==NULL)
			if(ecoli==NULL)
				reference=first;
			else
				reference=ecoli;
		reference_length=0;
		for(s=reference->sequence;*s!='\0';s++)
			if(base(*s))
				reference_length++;
	}

	/* get alignment_length */
    alignment_length=get_maximum_length(orgs);

	/* read column, sequence specification */
	sequence_specified=column_specified=0;
	sequence_error=column_error=0;
	more=1;
	for(i=1;i<argc && more;i++)
	{
		if(argv[i][0]=='-')
		{
			if(compare(argv[i], "-help"))
				/* dealt with before */;
			else if(compare(argv[i], "-pack"))
				/* dealt with before */;
			else if(compare(argv[i], "-nopack"))
				/* dealt with before */;
			else if(argv[i][1]=='c')
			{
				/* columns */

				static int len, j;
				static int error;

				if(column_specified)
				{
					fprintf(stderr, "ERROR--Columns specifed twice\n");
					printhelp(0, argv[0]);
					exit(1);
				}
					
				len=strlen(argv[i]);
				for(j=0;j<len;j++)
					if(argv[i][j]=='=')
						break;
				if(j==len)
				{
					fprintf(stderr,
						"ERROR--No columns specified on -c option\n");
					printhelp(0, argv[0]);
					exit(1);
				}
				if(argv[i][j+1]=='\0')
				{
					fprintf(stderr,
						"ERROR--No columns specified on -c option\n");
					printhelp(0, argv[0]);
					exit(1);
				}
					
				error=read_pair(&argv[i][j+1], &col_start, &col_end,
					1, alignment_length);
				if(error)
				{
					fprintf(stderr,
						"ERROR--Error reading column specification\n");
					printhelp(0, argv[0]);
					exit(1);
				}
				column_specified=1;
			}
			else if(argv[i][1]=='p')
			{
				/*position*/

				static int len, j;
				static int error;

				if(sequence_specified)
				{
					fprintf(stderr,
						"ERROR--Sequence positions specifed twice\n");
					printhelp(0, argv[0]);
					exit(1);
				}

				len=strlen(argv[i]);
				for(j=0;j<len;j++)
					if(argv[i][j]=='=')
						break;
				if(j==len)
				{
					fprintf(stderr,
						"ERROR--No list specified on -p option\n");
					printhelp(0, argv[0]);
					exit(1);
				}
				if(argv[i][j+1]=='\0')
				{
					fprintf(stderr,
						"ERROR--No list specified on -p option\n");
					printhelp(0, argv[0]);
					exit(1);
				}
					
				error=read_pair_list(&argv[i][j+1], seq_starts, seq_ends,
					1, reference_length, "Sequence Position list");
				if(error)
				{
					printhelp(0, argv[0]);
					exit(1);
				}
				sequence_specified=1;
			}
			else if(argv[i][1]=='r')
			{
				/*reference*/
				/* dealt with before */
			}
			else
			{
				fprintf(stderr,
					"ERROR--Invalid option: %s\n", argv[i]);
				printhelp(0, argv[0]);
				exit(1);
			}
				
		}
		else
		{
			/* files: dealt with before */
		}
	}

	if(!reconfigure_alignment(seq_starts, seq_ends,
		seq_col_starts, seq_col_ends,
		col_start, col_end,
		sequence_specified, column_specified,
		orgs, reference, &alignment_length, &mask))
	{
		fclose(in);
		fclose(out);
		if(log!=stderr)
			fclose(log);
		fclose(genbank);
		exit(0);
	}

	if(packing)
	{
		list start, end;

		/* remove common gaps, by filling mask with zeros */
		for(start=firstnode(seq_col_starts),end=firstnode(seq_col_ends);
				start!=NULL;
				start=nextnode(start),end=nextnode(end))
			removegaps(orgs, mask,
				thisint(start), thisint(end),
				alignment_length);
	}


    calc_num_base_pairs(orgs, mask, alignment_length);

    startlist(orgs);
    while((taxa=listnext(orgs))!=NULL)
	    printentry(taxa, mask, alignment_length, out);	

    /* all done, so close files and quit */
    fclose(in);
    fclose(out);
    if(log!=stderr)
	fclose(log);
    fclose(genbank);
    exit(0);
}

printlist(o)
list o;
{
    list tmp=listnode(o);
    organism x;
    startlist(o);
    while((x=listnext(o))!=NULL)
    {
	    fprintf(stderr, "%s\n", x->name);
    }
    setnode(o, tmp);
}

printseqs(o)
list o;
{
    list tmp=listnode(o);
    organism x;
    startlist(o);
    while((x=listnext(o))!=NULL)
    {
	    fprintf(stderr, "%s:%s\n", x->name, x->sequence);
    }
    setnode(o, tmp);
}
